# -----------------------------------------------------------------------------
# This is all the main game logic!
#
# Possibly too much in here, but it never really felt out of control, so
# I haven't moved any more out.
#
# v1.00 Apr 09 2005 Public release
# V1.10 May 11 2006 Add much more complete skinning - but not text yet
# v1.11 May 12 2006 Move [appearance] into skins
#                   Add comments for menu items
# v2.00 Aug 22 2008 Finish skinning with all text/menu items
#                   Require at least python 2.4
#
# Copyright 2005-2008 Ron Dippold
# -----------------------------------------------------------------------------

VERSION = "2.00"

import pygame
from pygame.locals import *

import ConfigParser
import os
import random

from Wave import Wave
from Ship import Ship, ShipCount
from TextSprite import *
from Missile import Sam, Sam_f, BaseMissile, Bullet, Bomb, Cruise, Cruise_f, Bomber, Mine
from Plane import Plane, Copter, Fighter
from Boom import Boom
from Cloud import Cloud
from Score import Score
from MySound import MySound
from Menu import Menu
import Music

import Demo
import Util
import TimedEvent

# -----------------------------
# Config
# -----------------------------

# Screen sizes - all fixed right now - if you change, strange things happen
GAME_SIZE = ( 640,480 )
PLAY_RECT = pygame.Rect( 0, 0, 512, 480 )
UI_RECT   = pygame.Rect( 512, 0, 128, 480 )

# base speed of the waves, ship, etc. Pixels/sec
BASE_SPEED = 80


# Types of app FOCUS
FOCUS_OFF, FOCUS_LAZY, FOCUS_ACTIVE = range(1,4)

# --------------------------------------------
# Globals (to this module)
# Technically this is evil but it works well
#  enough for something this simple
# --------------------------------------------

# This holds the lists of sprites - it's just a public struct
class sprite_set:
	ship            = None # our ship, and any sams

	planes			= None # Planes in formation
	bullets			= None # non-shootable enemy bombs/bullets
	
	harmless        = None # Harmless stuff like flames
	text            = None # onscreen text
	clouds			= None # Fluffy clouds in background

	demo            = None # demo mode stuff
	menu			= None # onscreen menus

# Appearance items
class skin:
	choices			= {}   # { skin_name : skin_file }
	name			= "Default" # name of skin
	
	# general
	smooth_waves	= True
	show_divider	= True

	# Colors - see skins/Default.txt for what all these mean
	class color:
		sky				= ( (0,0,40), )
		waves			= ( (0,0,175), (80,80,255) )
		fighters		= ( (200,0,0), (0,200,0), (0,0,200), (200,0,200),
							(0,200,200), (200,200,200) )

		text_logo		= ( (255,255,255), (255,0,0), (100,100,255) )
		text_gameover	= ( (255,0,0), (80,0,0), (0,0,160) )
		text_rank		= ( (80,80,255), (255,255,255), (160,160,255 ) )

		text_gameover1  = ( (255,255,255), )
		text_gameover2  = ( (255,255,255), )
		text_gameover3  = ( (160,160,160), )

		text_level1		= ( (0,255,255), )
		text_level2		= ( (255,255,255), )
		text_level3		= ( (160,160,160), )
		text_level4		= ( (255,255,255), )
		text_level5		= ( (255,100,0), )

		menu_text		= ( (255,255,255), )
		menu_bar		= ( (255,0,0), )
		menu_comment	= ( (255,255,0), )

		divider			= ( (160,160,160), )


# Other sprites that we need to track individually
class sprite:
	ship			= None # our ship sprite
	
	sam				= None # our ship's sam sprite
	sam_f			= None # flame for sam

	cruise			= None # cruise missile 
	cruise_f		= None #   flame for cruise
	cruise_time		= 0.0  # How long since we shot one

	bomber			= None # bomber
	bomber_time		= 0.0  # how long since we shot one

	paused_text		= None # 'PAUSED'

# Things we know about the planes
class planes:
	matrix			= None # (col,row) matrix of planes
	config			= None # ConfigParser of sprite/plane info
	patrol_time		= 0.0  # last time we started a patrol
	rows			= 0    # rows in matrix
	cols			= 0    # cols in matrix
	count			= 0    # number of planes left alive

# Info about the current level
class level:
	number			= 0		# numeric - 1 for level 1, etc.
	lastloaded		= 0		# number for last successfully loaded level
	
	difficulty		= 1     # numb
	probability		= 0.0	# probability that various bad things will happen

	done			= False # Level over?

	planes			= []   # setup list of all planes in level
	# <...> 			   # various other options from config file

# overall game state 
class game:
	STATE_GAME, STATE_OVER, STATE_DEMO = range(3)
	state			= None

	ship_count		= None	# Sprite holding number of ships in stock
	score			= None  # sprite holding current score

	ranks			= {}	# Rank names, like { 6: 'Gunner' }
	difficulties	= {}	# difficulty names for 1, 2, 3

	ocean_sound		= None  # Background waves sound

	screen 			= None # Pygame display screen
	background		= None # Background surface

	focus			= FOCUS_ACTIVE # do we have app focus?

	main_menu		= []   # Main menu
	ingame_menu		= []   # menu in game

	keys			= [K_LEFT, K_RIGHT, K_SPACE] # [left, right, fire]
	directories		= {}   # { "sprites":..., "sounds":..., "music":..., "levels":... }

# -----------------------------------------------------------------------------
#
# Overall game states
#
# -----------------------------------------------------------------------------

def game_over():
	"""
	All our guys are dead, game over!
	"""
	game.state = game.STATE_OVER

	if game.music:
		Music.play( game.directories["music"], "yohoho.wav" )

	# But wait two seconds for it to sink in
	TimedEvent.add( 1.5, game_over_showtext_callback )


def game_over_showtext_callback( _seconds, _data ):
	"""
	Display 'game over!' text.
	"""

	# 'Your fleet has been sunk!'
	y = PLAY_RECT.height/2 - 40
	text = TextSprite( "Your fleet has been sunk!", 36, skin.color.text_gameover1[0], (256,y ) )
	sprite_set.text.add( text )

	# 'GAME OVER'
	y = text.rect.bottom + 10
	text = WavyText( "GAME OVER", 48, skin.color.text_gameover[0], skin.color.text_gameover[2], 
		(256,y), 10, 3, skin.color.text_gameover[1], 2)
	sprite_set.text.add( text )
	y = text.rect.bottom + 20

	# show it for a few seconds
	TimedEvent.add( 1.0, game_over_ranktext_callback, _data=y )

def game_over_ranktext_callback( _seconds, y ):
	"""
	Add 'You reached the rank of'
	"""

	# 'You reached the rank of'
	text = TextSprite( "You reached the rank of", 36, skin.color.text_gameover2[0], (256,y) )
	sprite_set.text.add( text )
	y = text.rect.bottom+10

	# 'GALLEY SLAVE'
	text = WavyText( game.rank.upper(), 48, skin.color.text_rank[0], skin.color.text_rank[2], (256,y), -20,
					 _waveheight=5, _colormid=skin.color.text_rank[1], _midheight=4 )
	sprite_set.text.add( text )
	y = text.rect.bottom

	text = "Difficulty: %d - %s" % (game.difficulty, game.difficulties[game.difficulty])
	text = TextSprite( text, 24, skin.color.text_gameover3[0], (256,y) )
	sprite_set.text.add( text )
	y = text.rect.bottom+10

	# Show a few more seconds
	TimedEvent.add( 4.0, game_over_killtext_callback, y )

def game_over_killtext_callback( _seconds, _data ):
	"""
	Kill 'Game over!' text, wait 2 seconds
	"""
	sprite_set.text.empty()

	TimedEvent.add( 2.0, game_over_demomode_callback )

def game_over_demomode_callback( _seconds, _data ):
	"""
	Put game in demo mode
	"""
	game_done()


def game_start():
	"""
	Start a new game!
	"""

	# kill whatever demo we're in
	demo_stop()

	# Note that we're starting game...
	game.state = game.STATE_GAME

	# reset the ship position and number of ships
	sprite.ship.reset( PLAY_RECT.width/2 )
	sprite.ship.direction_reset()
	game.ship_count.set_count(2)

	# reset the score, and set score multiplier for difficulty level
	game.score.reset()
	game.score.set_multiplier( { 1:1, 2:1.5, 3:2.0 }[game.difficulty] )

	# our rank is minimal
	game.rank = game.ranks[1]

	# initialize the first level
	level_init( level_get_config( 1 ) )
	#print "STARTING AT HIGHER LEVEL"
	#level_init( level_get_config( 5 ) )

	# unmute sound in case it's muted, and start the background loop
	MySound.unmute()
	game.ocean_sound.play(-1)

def game_done():
	game.hiscore.write( "hiscore" )
	demo_start()


# -----------------------------------------------------------------------------
#
# Clouds
#
# -----------------------------------------------------------------------------
def clouds_init():
	"""Init the cloud static stuff"""
	Cloud.set_clip( PLAY_RECT )
	Cloud.set_WaveHelper( Wave.Helper )
	sprite_set.clouds = pygame.sprite.RenderUpdates()
	Cloud.preload()
	
def clouds_add():
	"""Add clouds if there aren't some already"""
	if len(sprite_set.clouds)==0:
		for i in range(3):
			sprite_set.clouds.add( Cloud() )

def clouds_clear():
	"""clear all clouds off screen"""
	sprite_set.clouds.empty()


# -----------------------------------------------------------------------------
#
# Demo stuff
#
# -----------------------------------------------------------------------------

def demo_start():
	# Clear level, turn off sound, set wave to nominal.
	MySound.mute()
	level_clear()
	level_set_wave_speed( 1 )

	# Go into demo state
	game.state = game.STATE_DEMO
	Demo.start( sprite_set.demo, PLAY_RECT )

def demo_stop():
	Demo.stop()

# -----------------------------------------------------------------------------
#
# Initialize, clear a level
#
# -----------------------------------------------------------------------------

def level_get_config( _level ):
	"""
	Return a SafeConfigParser() for the given level file, or None
	if there is no file.
	"""

	# Our level number is the level we wanted to read
	level.number = _level

	# Try reading it - if successful, get the config
	file = os.path.join( game.directories["levels"], "%003d.txt" % _level )
	if os.path.isfile(file):
		config = ConfigParser.SafeConfigParser()
		config.read( file )
		return config
	else:
		return None


def level_clear():
	# ----------------------------------------
	# Clear out/create the lists
	# ----------------------------------------

	sprite_set.bullets.empty()
	sprite_set.planes.empty()
	sprite_set.harmless.empty()
	sprite_set.text.empty()

	# If these exist, get rid of them
	sprite.cruise, sprite.cruise_f, sprite.cruise_time = ( None, None, 0.0 )
	sprite.bomber, sprite.bomber_time = ( None, 0.0 )

def level_set_wave_speed( _lvl ):
	# Figure out the wave speed
	speed = BASE_SPEED
	if _lvl <3:
		speed /= 2
	if not ( _lvl %2):
		speed *= -1
	Wave.new_speed( speed )


def level_init( _config ):
	"""
	Initialize a level, given a ConfigParser() for the settings.
	or None for a blank level
	"""


	# ----------------------------------------
	# Parse the various config file options
	# ----------------------------------------

	# In case we need it again
	level.config = _config

	# level difficulty
	level.difficulty = (level.number+1) * game.difficulty
	
	if _config:
		level.mines = min(_config.getint('enemies','mines'),2)
		level.cruise = _config.getboolean('enemies','cruise')
		level.bomber = _config.getboolean('enemies','bomber')
	else:
		level.mines = 0
		level.cruise = False
		level.bomber = False

	level.planes = []
	if _config:
		for i in "123456790":
			if _config.has_option( 'planes', i ):
				level.planes += [ _config.get( 'planes', i ) ]
			else:
				break

	# Clear the level
	level_clear()

	# ----------------------------------------
	# Now add stuff back in
	# ----------------------------------------

	# create/add all the planes
	planes_create( PLAY_RECT, level.planes )

	# create some mines if we need to
	if level.mines:
		mines_create( level.mines )

	# Set the wave speed for this level
	level_set_wave_speed( level.number )

	# The level is not done yet
	level.done = False

	# Any outstanding events... never mind
	TimedEvent.remove_all()


# -----------------------------------------------------------------------------
#
# All the stuff for a level being done
#
# -----------------------------------------------------------------------------

def level_check_for_done():
	"""
	Check if the level should be done, end it if it is
	"""

	# If there are no planes, cruise missiles, or bombers left, we're done!
	if not planes.count and not sprite.cruise and not sprite.bomber:
		level.done = True

		# If we killed an idling plane, just do the standard end
		if planes.killed_idle:
			# Show the level done text in two seconds
			TimedEvent.add( 2.0,level_done_showtext_callback )

		# otherwise, party time!
		else:
			TimedEvent.add( 1.0, level_done_rampage_callback )

RPOS = (256,140)
def level_done_rampage_callback( _seconds, _info ):
	"""
	All the planes we killed were patrolling!
	"""

	# Bonus score
	game.score.add_score( min(level.number,10) * 500 )

	# 'RAMPAGE'
	text = TextSprite( "RAMPAGE!", 64, skin.color.text_level5[0], RPOS )
	sprite_set.text.add( text )
	y = text.rect.bottom+20

	# Show a bunch of explosions, then continue as normal
	TimedEvent.add( 0.1, level_done_rampage2_callback, _count=40 )
	TimedEvent.add( 4.0, level_done_showtext_callback )

def level_done_rampage2_callback( _seconds, _info ):

	x = random.randrange( RPOS[0]-100, RPOS[0]+100 )
	y = random.randrange( RPOS[1]-100, RPOS[1]+100 )
	vx, vy = x-RPOS[0], y-RPOS[1]
	explosion_create( Boom.SIZE_MEDIUM, (x,y), (vx,vy), _life=1.0 )

def level_done_showtext_callback( _seconds, _info ):
	"""
	Show level complete text
	"""

	# Clear the level
	level_clear()

	# No leftover text
	sprite_set.text.empty()

	# 'CONGRATULATIONS'
	y = 60
	text = TextSprite( "CONGRATULATIONS!", 48, skin.color.text_level1[0], (256,y) )
	sprite_set.text.add( text )
	y = text.rect.bottom+20

	# The maximum rank we can get is determined by difficulty setting
	idx = len(game.ranks) - 3 + game.difficulty
	idx = min( idx, level.number+1 )
	idx = max( idx, 1 )
	rank = game.ranks[idx]

	# 'you have been promoted to' / 'you remain at'
	if game.rank!=rank:
		promotion = True
		text = TextSprite( "You have been promoted to", 24, skin.color.text_level2[0], (256, y) )
		game.rank = rank
	else:
		promotion = False
		text = TextSprite( "You remain at", 24, skin.color.text_level2[0], (256, y) )
	sprite_set.text.add( text )
	y = text.rect.bottom
	
	# 'BOATSWAIN'
	text = WavyText( game.rank.upper(), 48, skin.color.text_rank[0], skin.color.text_rank[2], (256,y+18), -20,
					 _waveheight=5, _colormid=skin.color.text_rank[1], _midheight=4 )
	sprite_set.text.add( text )

	# If we didn't get a promotion, add a helper line
	if not promotion:
		y = text.rect.bottom+5
		if game.difficulty<3:
			msg = "Try playing at a higher difficulty!"
		else:
			msg = "You've pretty much beaten the game!"
		text = TextSprite( msg, 24, skin.color.text_level3[0], (256, y) )
		sprite_set.text.add( text )
		
	y = text.rect.bottom+20

	# play a random tune
	if game.music:
		Music.play_random( game.directories["music"] )

	# show it for 2 seconds
	TimedEvent.add( 2.0, level_done_bonustext_callback, _data=y )


def level_done_bonustext_callback( _seconds, y ):
	"""
	Add bonus for remaining ships
	"""

	text = TextSprite( "Bonus! 50x", 36, skin.color.text_level4[0], (256,y) )
	sprite_set.text.add( text )
	y = text.rect.bottom

	count = ShipCount( sprite.ship, 0, pygame.Rect( 10, y, PLAY_RECT.width-20, 40 ),
					   _docenter = True)
	sprite_set.text.add( count )
	y = text.rect.bottom+20

	# Get the wave speed going
	level_set_wave_speed( level.number+1 )

	# For each ship, blink it in
	TimedEvent.add( 0.25, level_done_bonusships_callback, _count=game.ship_count.get_count(), \
					_data={ 'count':count, 'n':0 } )

def level_done_bonusships_callback( _seconds, _dict ):
	"""
	Add another ship to the list
	"""

	# Add one ship icon
	_dict['n'] += 1
	_dict['count'].set_count( _dict['n'] )

	# add 50 to score
	game.score.add_score( 50 )

	# are we done?
	if _dict['n']>=game.ship_count.get_count():
		TimedEvent.add( 2.0, level_done_killtext_callback)
		game.ship_count.increment()
	
def level_done_killtext_callback( _seconds, _info ):
	"""
	Kill level complete text
	"""
	sprite_set.text.empty()

	TimedEvent.add( 2.0, level_done_nextlevel_callback )

def level_done_nextlevel_callback( _seconds, _info ):
	"""
	Start new level
	"""

	# try reading the next level configuration
	newconfig = level_get_config( level.number+1 )
	if newconfig:
		level.config = newconfig

	# and init the level
	level_init( level.config )

# -----------------------------------------------------------------------------
#
# main initialization
#
# -----------------------------------------------------------------------------

import pygame.locals
def key_read( _config, _key, _default ):
	"""
	Read a key value from the ConfigParser() - make sure it's a
	pygame.constants value, and return it. Otherwise return _default.
	"""

	value = _default
	if _config.has_option( "keys", _key ):
		key = _config.get( "keys", _key )
		try:
			value = eval("pygame.locals."+key)
		except AttributeError:
			pass
	return value

def directory_read( _config, _key ):
	"""
	Read a directory setting from the ConfigParser().
	If none found, the _key IS the default as well.
	"""
	value = _key
	if _config.has_option( "directories", _key ):
		value = _config.get( "directories", _key )
	return os.path.join( "skins", value )

def map_value( _dict, _value, _default ):
	"""
	Find _value in _dict of { string : value }, return associated value.
	If not found, return _default.
	Case will be ignored!
	"""
	# add entries for lower case
	for key,val in _dict.items():
		_dict[key.lower()] = val

	return _dict.get( _value.lower(), _default )


def read_colors( _config, _key, _size, _default=None ):
	"""
	Parse "(R,G,B) (R,G,B)" for up to _size tuples (if _size set).
	returns real [ (r,g,b), (r,g,b) ] list of tuples, or
	   _default if it failed.
	"""

	# Grab it from the config, default if it doesn't exist
	if not _config.has_option( "colors", _key ):
		return _default
	args = _config.get( "colors", _key )
		
	# chop it up into different bits
	split = args.split("(")
	colors = []
	for rgb in split:
		rgb = rgb.replace(')','').strip()
		if not rgb: continue
		# Pull out into ( r,g,b )
		rgb = rgb.split(',')[:3]
		if len(rgb)<3: continue
		# convert to integers and clamp from 0-255
		try:
			rgb = [ min(max(int(c),0),255) for c in rgb ]
		except ValueError:
			continue
		colors.append( tuple(rgb) )

	# if we got nothing, then return default
	if not colors:
		return _default

	# trim down to side or expand up to size
	if _size:
		colors = colors[:_size]
		if len(colors)<_size:
			colors += colors[-1] * (size-len(colors))

	# tuple it and return
	return tuple(colors)
		

def read_user_config():
	# Read the user configuration
	config = ConfigParser.SafeConfigParser()
	config.read( "user_config.txt" )

	# get the skin file to use
	skin.name = "Default"
	if config.has_option( "skin", "skin" ):
		skinname = config.get( "skin", "skin" )
		if skinname in skin.choices:
			skin.name = skinname
	SKINS_CHOICE[ 'choice' ] = skin.name
		
	# read key bindings
	game.keys[0] = key_read( config, "left", game.keys[0] )
	game.keys[1] = key_read( config, "right", game.keys[1] )
	game.keys[2] = key_read( config, "fire", game.keys[2] )

	# Get difficulty, clamp it to 1-3
	game.difficulty = min(max(config.getint( "menus", "difficulty" ),1),3)

	# Read clouds, music, display settings - configure menus appropriately
	boolset =  { "on" : True, "off": False,
				 "true" : True, "false" : False,
				 "1" : True, "0": False }
	game.clouds = map_value( boolset, config.get( "menus", "clouds" ), True )
	CLOUDS_CHOICE["choice"] = { True: "On", False: "Off" }[game.clouds]

	game.music = map_value( boolset, config.get( "menus", "music" ), True )
	MUSIC_CHOICE["choice"] = { True: "On", False: "Off" }[game.music]
	
	game.fullscreen = map_value( { "window" : False, "fullscreen" : True }, \
				  config.get( "menus", "display" ), False )
	DISPLAY_CHOICE["choice"] = { True: "Fullscreen", False: "Window" }[game.fullscreen]

def read_skin_config( _skin ):

	config = ConfigParser.SafeConfigParser()
	config.read( skin.choices[_skin] )
	game.directories["levels"] = directory_read( config, "levels" )
	game.directories["sprites"] = directory_read( config, "sprites" )
	game.directories["sounds"] = directory_read( config, "sounds" )
	game.directories["music"] = directory_read( config, "music" )

	skin.smooth_waves = config.getboolean( 'appearance', 'smooth_waves' )
	skin.show_divider = config.getboolean( 'appearance', 'show_divider' )

	# We have so many colors do it proceduraly
	for item in (
		"sky:1", "waves:2", "fighters", "divider", "menu_text", "menu_bar", "menu_comment",
		"text_logo:3", "text_gameover:3", "text_rank:3",
		"text_gameover1", "text_gameover2", "text_gameover3",
		"text_level1", "text_level2", "text_level3", "text_level4", "text_level5",
		 ):
		item = item.split(":")
		color = item[0]
		try:
			count = int(item[1])
		except:
			count = None
		setattr( skin.color, color,
				 read_colors( config, color, count, getattr( skin.color, color ) ) )

def init():
	"""
	Initialize the display and the sprite lists

	"""

	# get which skins exist - { name : filename }
	skin.choices = {}
	comments = {}
	config = ConfigParser.SafeConfigParser()
	for fname in os.listdir( "skins" ):
		# ignore extra/backup files
		if fname.startswith(".") or fname.endswith("~"):
			continue
		fullname= os.path.join( "skins", fname )
		# ignore directories		
		if not os.path.isfile( fullname ):
			continue

		# Get name/description of skin
		config.read( fullname )
		if config.has_option( "info", "name" ):
			fname = config.get( "info", "name" )
		else:
			fname = fname.split(".")[0]

		if config.has_option( "info", "desc" ):
			comments[fname] = config.get( "info", "desc" )
		else:
			comments[fname] = ""

		# add to list of choices
		skin.choices[ fname ] = fullname

	if not skin.choices:
		print "No skin files found in skins directory."
		return False
	
	SKINS_CHOICE[ "choices" ] = skin.choices.keys()
	SKINS_CHOICE[ "comments" ] = [ comments[key] for key in SKINS_CHOICE["choices"] ]

	# Read the user's configuration
	read_user_config()

	#
	# Init pygame
	#
	pygame.init()

	MySound.init()

	# Create the display either windowed or fullscreen
	if game.fullscreen:
		flags = FULLSCREEN
	else:
		flags = 0
	game.screen = pygame.display.set_mode( GAME_SIZE, flags )

	# Make sure we're in the right mode for now
	if pygame.display.get_surface().get_size()!=GAME_SIZE:
		print "Sorry, this only runs in %dx%d mode" % GAME_SIZE
		return False

	# always hide the mouse!
	pygame.mouse.set_visible(False)

	# We don't need these
	pygame.event.set_blocked( ( MOUSEMOTION, MOUSEBUTTONUP, MOUSEBUTTONDOWN,
								JOYAXISMOTION, JOYBALLMOTION, JOYHATMOTION,
								JOYBUTTONUP, JOYBUTTONDOWN,
								VIDEORESIZE ) )

	# Make a background, load background.gif if it exists
	game.background = pygame.Surface( game.screen.get_size() ).convert()

	new_skin( skin.name )

	# Read the high score, if any
	game.hiscore.read( "hiscore" )
	
	return True


def new_skin( _skin ):
	skin.name = _skin
	read_skin_config( _skin )

	# Read the configuration for the graphics, level names
	config = ConfigParser.SafeConfigParser()
	file = os.path.join( game.directories["sprites"], "CONFIG.txt")
	if not os.path.isfile( file ):
		# This should raise file not found
		f = open( file )
	else:
		config.read( file )

	# Get the sound config
	sound_config = ConfigParser.SafeConfigParser()
	file = os.path.join( game.directories["sounds"], "CONFIG.txt")
	if not os.path.isfile( file ):
		# This should raise file not found
		f = open( file )
	else:
		sound_config.read( file )

	# Tell the image loader where our data is at and our configuration info
	Util.loader_config( game.directories, config, sound_config  )

	# Get level names
	i = 1
	game.ranks = {}
	while config.has_option( 'ranks', str(i) ):
		game.ranks[i] = config.get( 'ranks', str(i) )
		i += 1

	# Get difficulty names
	game.difficulties = {}
	for i in range(1,4):
		game.difficulties[i] = config.get( 'difficulty', str(i) )
	difficulty_choice_init( game.difficulty )

	try:
		bgimage = Util.image_loader ('background.gif' )
		game.background.blit( bgimage, (0,0) )
		del bgimage
	except pygame.error:
		game.background.fill( skin.color.sky[0] )
		

	# We'll use this quite a bit - center of the UI panel
	ui_x = UI_RECT.left+2+(UI_RECT.width-2)/2
	ui_y = 25

	# Add the play/score separator
	if skin.show_divider:
		pygame.draw.line( game.background, skin.color.divider[0], (UI_RECT.left,0),
						  UI_RECT.bottomleft, 2 )

	# Just drop the title directly on the background.
	# update() is to move it over a little from the 0 position
	# 'WAVY'
	Title = WavyText( "WAVY", 48, skin.color.text_logo[0], skin.color.text_logo[2], (ui_x,ui_y), 10,
					  2, skin.color.text_logo[1] )
	Title.update( 3 )
	game.background.blit( Title.image, Title.rect.topleft )
	ui_y = Title.rect.bottom+5

	# 'NAVY'
	Title = WavyText( "NAVY", 48, skin.color.text_logo[0], skin.color.text_logo[2], (ui_x, ui_y),
					  10, 2, skin.color.text_logo[1] )
	Title.update( 1.5 )
	game.background.blit( Title.image, Title.rect.topleft )
	ui_y = Title.rect.bottom+5
	del Title

	# Draw the background to screen and show it, since we never have any more
	# full screen updates
	game.screen.blit( game.background, (0,0) )
	pygame.display.update()

	# create a fixed width font for scoring, figure out how big it is, place it
	game.score = Score( (ui_x, ui_y) )
	# Create a high score down at the bottom
	game.hiscore = Score( (ui_x, UI_RECT.bottom-10 ), _color=(160,160,160),
						  _points=24, _digits=8 )

	# Set up the the waves
	Wave.setup( skin.smooth_waves, PLAY_RECT, _xspeed=0, _color=skin.color.waves,\
				_bgcolor=game.background.get_at( (0, PLAY_RECT.bottom-1) ) )

	# initialize the cloud static stuff
	clouds_init()
	if game.clouds:
		clouds_add()

	# Add the ship
	sprite_set.ship = pygame.sprite.RenderUpdates()
	sprite.ship = Ship( BASE_SPEED, PLAY_RECT.width/2, PLAY_RECT.width,
						Wave.Helper, game.keys )
	sprite_set.ship.add( sprite.ship )

	# Set up the ship count in the lower right
	countrect = pygame.Rect( UI_RECT.left+5, UI_RECT.bottom-95,
							 UI_RECT.width-7, 75 )
	game.ship_count = ShipCount( sprite.ship, 3, countrect )

	# Set the missile limits
	BaseMissile.set_clip( PLAY_RECT )
	BaseMissile.set_base_speed( BASE_SPEED )
	BaseMissile.set_WaveHelper( Wave.Helper )

	# set up our sprite sets
	sprite_set.planes = pygame.sprite.RenderUpdates()
	sprite_set.bullets = pygame.sprite.RenderUpdates()
	sprite_set.harmless = pygame.sprite.RenderUpdates()
	sprite_set.text = pygame.sprite.RenderUpdates()
	sprite_set.demo = pygame.sprite.RenderUpdates()
	sprite_set.menu = pygame.sprite.RenderUpdates()

	# Create a background ocean sound, play forever
	game.ocean_sound = Util.sound_loader( 'ocean' )

	# Preload the graphics for most of the sprites
	Bomb.preload()
	Bomber.preload()
	Boom.preload()
	Copter.preload()
	Cruise.preload()
	Cruise_f.preload()
	Fighter.preload( skin.color.fighters )
	Sam.preload()
	Sam_f.preload()
	# no Ship() preload, already did it

	# Load up our menus for when we need them
	main_menu_create()
	ingame_menu_create()

	# Init a cleared level, to get some stuff setup
	level_init( None )
	demo_start()

	

def kill_skin():
	"""unload old skin stuff"""
	Ship.unload()
	
# -----------------------------------------------------------------------------
#
# Turn sound on and off temporarily
#
# -----------------------------------------------------------------------------
def sound_pause():
	"""Pause sound and music when we lose focus or pause"""
	if pygame.mixer:
		pygame.mixer.pause()
	Music.pause()

def sound_unpause():
	"""Resume sound and music"""
	if pygame.mixer:
		pygame.mixer.unpause()
	Music.unpause()

# -----------------------------------------------------------------------------
#
# Main menu
#
# -----------------------------------------------------------------------------

# We share this between both menus
MUSIC_CHOICE = {
	# Choose whether music is on or off
	"name": "music",
	"type": "choice",
		  
	"label": "Music",
	"choices": [ "On", "Off" ],
	"comments": [ "After-level fanfare" ] *2,
	"choice": "On"
}

def display_setting_callback( _menu, _item ):
	if _menu.get_choice( "display")=="Window":
		fullscreen = False
	else:
		fullscreen = True

	# Switch between windowed and fullscreen
	if game.fullscreen != fullscreen:
		game.fullscreen = fullscreen

		# Change the mode, blit the background to it, draw occasional stuff
		if game.fullscreen:
			flags = FULLSCREEN
		else:
			flags = 0
		game.screen = pygame.display.set_mode( GAME_SIZE, flags )
		game.screen.blit( game.background, (0,0) )
		draw_occasional_stuff( True )
		pygame.display.update()

		# Hide the mouse if it's fullscreen, show if it's windowed
		#pygame.mouse.set_visible( not fullscreen )

		# always hide the mouse!
		pygame.mouse.set_visible(False)


DISPLAY_CHOICE = {
	"name": "display",
	"type": "choice",

	"label": "Display",
	"choices": [ "Window", "Fullscreen" ],
	"choice": "Window",
	"action": display_setting_callback
	}


def skin_setting_callback( _menu, _item ):
	kill_skin()
	new_skin( _menu.get_choice("skin") )
	game.main_menu.activate()
	game.main_menu.set_cursor( SKINS_CHOICE )

SKINS_CHOICE = {
	"name": "skin",
	"type": "choice",

	"label": "Skin",
	"choices": [ "Default" ],
	"choice": "Default",
	"action": skin_setting_callback
	}

def cloud_setting_callback( _menu, _item ):
	game.clouds = ( _menu.get_choice("clouds") == "On" )
	if game.clouds:
		clouds_add()
	else:
		clouds_clear()

CLOUDS_CHOICE = {
	"name": "clouds",
	"type": "choice",

	"label": "Clouds",
	"choices": [ "On", "Off" ],
	"comments": [ "Lazily drifting clouds" ] * 2,
	"choice": "On",
	"action": cloud_setting_callback
	}
	

def main_menu_cancel_callback( _menu, _item ):
	"""Never mind, go back to the demo"""
	game.main_menu.deactivate()

	# Set the difficulty, music setting
	game.difficulty = int( _menu.get_choice("difficulty")[0] )
	game.music = (_menu.get_choice("music") == "On")
	if not game.music:
		Music.stop()

def main_menu_start_callback( _menu, _item ):
	"""Start a new game"""

	# Do the same stuff as cancel, but then start the game
	main_menu_cancel_callback( _menu, _item )
	game_start()

def main_menu_exit_callback( _menu, _item ):
	"""Post a quite event to be handled later"""
	pygame.event.post( pygame.event.Event( QUIT ) )


DIFFICULTY_CHOICE = {
	# Choose level of difficulty, one of 3 choices, set with func
	"name": "difficulty",
	"type": "choice",
		  
	"label": "Difficulty",
	"choices": [""],
	"comments" : [ "Very easy - for beginners", "Now it's interesting", "Duck and cover!" ],
	"choice": ""
	}
def difficulty_choice_init( _difficulty ):
	# Turn the difficulties into "1 - Beginner", etc.
	diffs = []
	for val, name in game.difficulties.items():
		diffs += [ "%d - %s" % (val,name) ]
	diffs.sort()
	DIFFICULTY_CHOICE["choices"] = diffs
	DIFFICULTY_CHOICE["choice"] = diffs[_difficulty-1]

def main_menu_create():
	"""This creates the main menu but doesn't display it."""

	menu = 	[
		DISPLAY_CHOICE,
		MUSIC_CHOICE,
		CLOUDS_CHOICE,
		SKINS_CHOICE,
		DIFFICULTY_CHOICE,
		
		{ # Play!
			"name": "start",
			"type": "action",

			"label": "Start Game!",
			"action": main_menu_start_callback
			},
		{ # Never mind
			"name": "cancel",
			"type": "action",
			"keys": K_ESCAPE,

			"label": "Back to Demo",
			"action": main_menu_cancel_callback
			},
		{ # bye bye
			"name": "exit",
			"type": "action",

			"label": "Exit Wavy Navy",
			"action": main_menu_exit_callback
			}
		]
	game.main_menu = Menu( menu, PLAY_RECT,
						   _spriteset = sprite_set.menu, _startitem = "start",
						   _textcolor = skin.color.menu_text[0],
						   _barcolor = skin.color.menu_bar[0],
						   _commentcolor = skin.color.menu_comment[0]
						   )


# -----------------------------------------------------------------------------
#
# In-game menu
#
# -----------------------------------------------------------------------------


def ingame_menu_resume_callback( _menu, name ):
	"""Resume the game"""
	sound_unpause()
	game.music = (_menu.get_choice("music") == "On")
	if not game.music:
		Music.stop()
	game.ingame_menu.deactivate()
	sprite.ship.direction_reset()

def ingame_menu_quit_callback( _menu, _item ):
	"""Quit the game"""
	sound_unpause()
	game.ingame_menu.deactivate()
	game_done()

def ingame_menu_create():
	"""This creates the in-game menu but doesn't display it."""

	# Turn the difficulties into "1 - Beginner", etc.
	diffs = []
	for val, name in game.difficulties.items():
		diffs += [ "%d - %s" % (val,name) ]
	diffs.sort()
	
	menu = 	[
		DISPLAY_CHOICE,
		MUSIC_CHOICE,
		CLOUDS_CHOICE,

		{ # quit game
			"name": "quit",
			"type": "action",

			"label": "Quit Game",
			"action": ingame_menu_quit_callback
			},
		{ # resume game
			"name": "resume",
			"type": "action",
			"keys": K_ESCAPE,
		  
			"label": "Resume Game",
			"action": ingame_menu_resume_callback
		  },
		]
	game.ingame_menu = Menu( menu, PLAY_RECT,
							 _spriteset = sprite_set.menu,
							 _startitem = "resume",
							 _textcolor = skin.color.menu_text[0],
							 _barcolor = skin.color.menu_bar[0],
						     _commentcolor = skin.color.menu_comment[0]
							 )

	

# -----------------------------------------------------------------------------
#
# SAM logic
#
# -----------------------------------------------------------------------------
def sam_dead( _sam ):
	# no need to remove it from group, it kill()ed itself
	sprite.sam = None
	# But we need to remove the flame
	sprite.sam_f.kill()
	sprite.sam_f = None

def sam_fire( _pos ):
	"""
	Request to fire a missile
	"""

	# Can't if we've already fired one
	if sprite.sam:
		return

	sprite.sam = Sam( _pos, sam_dead )
	sprite_set.ship.add( sprite.sam )

	sprite.sam_f  = Sam_f( sprite.sam )
	sprite_set.ship.add( sprite.sam_f )

def sam_check_collisions():
	"""
	Look for our shots hitting anything
	"""
	# The True will invoke the plane killed callbacks, taking them out
	items = pygame.sprite.spritecollide( sprite.sam, sprite_set.planes, True )
	if items:
		# kill the missile itself, also takes out flame
		sprite.sam.kill()
		# add score!
		for item in items:
			game.score.add_score( item.score() )

# -----------------------------------------------------------------------------
#
# Ship extra logic
#
# -----------------------------------------------------------------------------

def ship_death_logic():
	"""
	Handle the ship is dying/ship is dead/ship is hit stuff
	"""
	#
	# We're completely dead - see if we have and ships left,
	# otherwise it's game over
	if sprite.ship.is_dead():
		if game.ship_count.get_count() > 0:
			game.ship_count.decrement()
			# This resets the dead/dying too
			sprite.ship.reset( PLAY_RECT.width/2 )
			# can't be killed for two seconds
			sprite.ship.set_invulnerable( 2.0 )
		else:
			game_over()
		return
			
	# If we're dying, don't do anything special	
	if sprite.ship.is_dying():
		return

	# Otherwise, check for any hits
	if not sprite.ship.is_invulnerable():
		items = pygame.sprite.spritecollide( sprite.ship, sprite_set.bullets, True )
		items += pygame.sprite.spritecollide( sprite.ship, sprite_set.planes, True )
		if items:
			# Make a big explosion
			explosion_create( Boom.SIZE_LARGE, sprite.ship.rect.center, (0,0) )
			# Tell the ship to go into death agonies
			sprite.ship.die( PLAY_RECT.bottom )
			# Have all planes return to idling
			planes_return_to_idle()

# -----------------------------------------------------------------------------
#
# Plane logic
#
# -----------------------------------------------------------------------------

def planes_create( _rect, _layout ):
	"""
	Create a new plane layout.
	_layout is a tuple of strings, one string per plane line.
	Each character is a space (no plane), C (copter) or F (fighter)
	"""

	# Figure out how many rows and columns by counting chars
	rows = len(_layout)
	cols = 0
	for row in _layout:
		cols = max( cols, len(row) )

	# to place our plane info so we can keep track
	planes.matrix = {}

	# tell our planes where the centers are and where the waves are
	Plane.set_wave_info( Wave.Helper, plane_hitwave_callback )

	# Set up our initial variables - these should be tunable later
	startx, starty = 40, 40
	deltax, deltay = 45, 25
	sizex = cols*deltax

	# x1 is where the left column would be when planes are in center -
	# if it's 90, then we want to march it from startx (=40) to
	# startx + (90-startx)*2 (=140)
	x1 = _rect.centerx - sizex/2
	Plane.set_idlerect( pygame.Rect( startx, starty, (x1-startx)*2, 20 ) )

	# Have planes fly in from far left at high speed
	Plane.flyin_from( (cols-1)*deltax )

	# set the level difficulty
	Plane.set_difficulty( level.difficulty, game.difficulty )

	for row in range(rows):
		for col in range(cols):
			char = _layout[row][col]
			offset = [ col*deltax, row*deltay ]
			if char in " .":
				continue
			elif char=='C':
				if row==1:
					plane = Copter( (col,row), (offset[0],offset[1]-deltay/2), plane_killed )
				else:
					plane = Copter( (col,row), offset, plane_killed )
			elif char=='F':
				plane = Fighter( (col,row), offset, plane_killed )
			else:
				print "*** Unknown plane type '%s' in '%s'" % ( char, _layout[row] )
				continue
			planes.matrix[(col,row)] = plane
			sprite_set.planes.add( plane )

	planes_set_playarea( _rect )
	planes.patrol_time = 0
	planes.patrol_total_time = 0
	planes.rows = rows
	planes.cols = cols

	# Track whether any killed planes were idling
	planes.killed_idle = False

	# Set the callback for copter firing a bullet
	Copter.set_bullet_callback( copter_fire_bullet )
	Fighter.set_bomb_callback( fighter_drop_bomb )

def planes_set_playarea( _rect ):
	"""
	Planes care about two rectangles - the 'in play' rectangle including
	the idling, and the 'patrol' rectangle, which is where they try to stay
	in while patrolling.
	"""

	# figure out where the very bottom plane is
	maxy = _rect.top
	planes.count = 0
	for plane in planes.matrix.values():
		if plane:
			planes.count += 1
			if plane.is_idle():
				maxy = max( maxy, plane.get_y() )
	maxy += 5

	# set the rects
	Plane.set_playarea( _rect, maxy )

def planes_return_to_idle():
	"""
	Send any patrolling planes back to idle.
	"""
	for plane in planes.matrix.values():
		if plane and not plane.is_idle():
			plane.return_to_idle()
	

def plane_killed( _plane ):
	"""A plane was shot down"""
	# remove the plane from the planes matrix
	planes.matrix[_plane.id] = None

	# Create an explosion - medium if idle, large if on patrol
	if _plane.is_idle():
		size = Boom.SIZE_MEDIUM
		planes.killed_idle = True
	else:
		size = Boom.SIZE_LARGE
	explosion_create( size, _plane.rect.center, [_plane.speed[0]*1.1,_plane.speed[1]-10] )
		
	# change the patrol rectangle in case this shrunk it
	planes_set_playarea( PLAY_RECT )


def plane_hitwave_callback( _plane, _pos ):
	"""Plane hit a wave - just make an explosion"""
	explosion_create( Boom.SIZE_MEDIUM, [_pos[0], _pos[1]], [0,0] )


def planes_start_patrol( _seconds ):
	"""Start Copters and Fighters going randomly"""

	# Add in the elapsed time
	planes.patrol_time += _seconds
	planes.patrol_total_time += _seconds

	# Max number of patrolling planes is the level difficulty plus
	#    one more for every so many seconds
	patrolmax = int(level.difficulty+planes.patrol_total_time/30)
	if Plane.patrol_count >= patrolmax:
		return

	# The longer we go, the more chance we have of a patrolling plane
	prob = level.difficulty * game.difficulty * planes.patrol_time * .003 * _seconds
	
	# Add in total patrol time - maxed out at 5% chance
	prob2 = (planes.patrol_total_time * .002) * _seconds
	prob2 = min( prob2, 0.05 )
	#print "%3d %.6f %.6f" % ( patrolmax, prob, prob2 )
	prob += prob2

	# Make sure there's always at least a 1% chance
	prob = max( prob, 0.01 )

	# If < prob (and any planes left) then start patrol
	if random.random()<prob and planes.count:
		planes.patrol_time = 0

		# pick a row to start with
		row = random.randrange( planes.rows )
		# Try looping all the rows in case this one is empty
		for i in range( planes.rows ):

			# look for an idling plane in this row, either from the
			#    left or from the right randomly
			plane = None
			if random.random()>0.5:
				idxrange = range( planes.cols-1, -1, -1 )
			else:
				idxrange = range( planes.cols )
			for col in idxrange:
				plane = planes.matrix.get( (col, row), None )
				if plane and plane.is_idle():
					break
				plane = None
			if plane:
				plane.patrol()
				break
			else:
				row = (row + 1) % planes.rows

def copter_fire_bullet( _plane, _pos, _speed ):
	"""
	Callback routine from Copter()
	_pos is initial position
	_speed is speed normalized to 0-1 in each dimension
	"""
	sprite_set.bullets.add( Bullet( _pos, _speed, bullet_hit ) )


def fighter_drop_bomb( _plane, _pos ):
	"""
	Callback routine from Fighter() - _pos is initial position
	"""
	sprite_set.bullets.add( Bomb( _pos, bullet_hit ) )

# -----------------------------------------------------------------------------
#
# Extra bullet stuff
#
# -----------------------------------------------------------------------------

def bullet_hit( _bullet ):
	"""
	Bullet (or bomb!) hit something, or died.
	"""
	# We only need a separate explosion if it hit the waves.
	if _bullet.hit_waves:
		explosion_create( Boom.SIZE_SMALL, _bullet.pos, [Wave.xspeed,0] )


def mines_create( _count ):
	if _count==1:
		x = 0
	else:
		x = Wave.size[0]/2
		
	y = Wave.Helper.max_top()
	for m in range( _count ):
		sprite_set.bullets.add( Mine( [x,y], Wave.xspeed, PLAY_RECT ) )
		x += Wave.size[0]


# -----------------------------------------------------------------------------
#
# Cruise missile
#
# -----------------------------------------------------------------------------

def cruise_start_patrol( _seconds ):
	"""
	IF there are cruise missiles on this level, and we don't have one yet, then
	launch one if our random roll is below the probability. The probability
	goes up over time till it's inevitable.
	"""

	if level.cruise and not sprite.cruise:
		sprite.cruise_time += _seconds

		# we increase this probability more slowly with time than planes
		prob = (20+level.difficulty) * sprite.cruise_time * .0015 * _seconds
		if random.random() <= prob:
			sprite.cruise_time = 0.0
			# Add a new cruise missile, randomly headed left or right
			cruise_add( True, random.choice((-1,1)) )

def cruise_add( _firstpass, _dir ):
	"""Add a cruise missile and its flame"""
	# Add the missile in the sprite_set.planes
	sprite.cruise = Cruise( _firstpass, _dir, cruise_dead )
	sprite_set.planes.add( sprite.cruise )

	# Add its flame in the harmless set
	sprite.cruise_f  = Cruise_f( sprite.cruise )
	sprite_set.harmless.add( sprite.cruise_f )

def cruise_dead( _cruise ):
	"""Cruise missile dead, remove it and flame. Maybe respawn."""
	# No need to cruise.kill(), it's already gone
	sprite.cruise = None
	sprite.cruise_f.kill()
	sprite.cruise_f = None
	sprite.cruise_time = 0.0

	# If we're 'dead' because we left the area
	if _cruise.hit_clip:
		# If that was the first pass, then make it go the other way along the bottom
		if _cruise.first_pass:
			cruise_add( False, -_cruise.dir )
	else:
		explosion_create( Boom.SIZE_LARGE, _cruise.rect.center, \
						  [_cruise.speed[0]*0.4,_cruise.speed[1]] )


# -----------------------------------------------------------------------------
#
# Bomber
#
# -----------------------------------------------------------------------------

def bomber_start_patrol( _seconds ):
	"""
	IF there are bombers on this level, and we don't have one yet, then
	launch one if our random roll is below the probability. The probability
	goes up over time till it's inevitable.
	"""
	if level.bomber and not sprite.bomber:
		sprite.bomber_time += _seconds

		# we increase this probability more slowly with time than planes
		prob = (20+level.difficulty) * sprite.bomber_time * .0005 * _seconds
		if random.random() < prob:
			sprite.bomber_time = 0.0

			# Add a new bomber from the opposite side of the screen that
			# the ship is on ( so it doesn't nail you instantly )
			dir = Util.sign( sprite.ship.get_xpos() - PLAY_RECT.centerx )
			if dir==0: dir = -1
			bomber_add( dir )

def bomber_add( _dir ):
	"""Add a bomber"""

	ypos = random.randrange( 50, Wave.Helper.min_top()-10 )
	distance = sprite.ship.rect.width * (3.0, 3.0, 2.25, 1.75)[game.difficulty]
	sprite.bomber = Bomber(
		_ypos = ypos,
		_dir = _dir,
		_bomb_distance = distance,
		_bomb_callback = fighter_drop_bomb,
		_dead_callback = bomber_dead )
	sprite_set.planes.add( sprite.bomber )

def bomber_dead( _bomber ):
	"""Bomber dead"""
	# No need to .kill(), it's already gone
	sprite.bomber = None
	sprite.bomber_time = 0.0

	# If we're 'dead' because we left the area, do nothing
	if _bomber.hit_clip:
		pass
	else:
		explosion_create( Boom.SIZE_LARGE, _bomber.rect.center, \
						  [_bomber.speed[0]*0.4,_bomber.speed[1]] )

# -----------------------------------------------------------------------------
#
# Explosions logic
#
# -----------------------------------------------------------------------------
def explosion_create( _size, _pos, _speed, _life=1.0 ):
	sprite_set.harmless.add( Boom( _size, _pos, _speed, _life ) )


# -----------------------------------------------------------------------------
#
# Handle focus tracking
#
# we used to use pygame.key.get_focused() but that was a bit iffy. So we
# do it the hard way and see where the window focus is by looking at the
# ACTIVEEVENT type pygame.events:
#
# 	state:1 gain:0 - mouse has left our active area
# 	state:1 gain:1 - mouse has reentered our active area
# 	state:2 gain:0 - another app has gained focus
# 	state:6 gain:1 - we have gained focus
# 	state:6 gain:0 - we've been minimized
#   
# -----------------------------------------------------------------------------

def focus_have_focus():
	"""
	Return True if we have focus and should be playing
	"""
	return game.focus in (FOCUS_ACTIVE, FOCUS_LAZY)

def focus_activeevent( _event ):
	"""
	Handle a pygame ACTIVEEVENT
	"""

	focus = game.focus

	# Another app got focus
	if _event.state==2 and _event.gain==0:
		focus = FOCUS_OFF
	# We were minimized
	elif _event.state==6 and _event.gain==0:
		focus = FOCUS_OFF
	# actively regained focus
	elif _event.state==6 and _event.gain==1:
		focus = FOCUS_ACTIVE

	# mouse wandered out of window
	elif _event.state==1 and _event.gain==0:
		if focus==FOCUS_ACTIVE:
			focus = FOCUS_LAZY
	# mouse wandered back in
	elif _event.state==1 and _event.gain==1:
		if focus==FOCUS_LAZY:
			focus = FOCUS_ACTIVE

	# nothing changed
	if focus==game.focus:
		return

	# game state is now new state
	game.focus = focus

	# We're now focused - resume game
	if focus_have_focus():
		# remove paused text
		sprite_set.text.remove( sprite.paused_text )
		# Ship is now no longer going left or right
		sprite.ship.direction_reset()
		# turn sound back on
		sound_unpause()
		# Make sure mouse isn't visible
		pygame.mouse.set_visible(False)
		# Redraw everything, it can act a little weird when we alt-tab back
		pygame.display.update()
		

	# We've lost focus - put up pause text
	else:
		if not sprite.paused_text:
			sprite.paused_text = BlinkText( "PAUSED", 48, ((255,255,255),(0,0,80)),
					   (256, game.screen.get_height()/2) )
		sprite_set.text.add( sprite.paused_text )

		# Turn off sound
		sound_pause()


# -----------------------------------------------------------------------------
#
# Draw sprites only occasionally
#
# -----------------------------------------------------------------------------
def sprite_draw_occasional( _sprite ):
	"""
	Draw a sprite only when it changes, not every frame. We assume that
	the sprite has not moved when we restore the background.
	Returns a changerect list.
	"""
			
	# We'll be using this a lot
	src_rect = _sprite.rect
			
	# Erase the old sprite
	game.screen.blit( game.background, src_rect.topleft, src_rect )
			
	# write the new sprite
	game.screen.blit( _sprite.image, src_rect.topleft )

	# return change box
	return [ src_rect ]


def draw_occasional_stuff( _force=False ):
	"""
	Update some stuff that we only draw when it actually changes,
	so we don't end up drawing it every frame.
	Returns a changelist.
	"""
	changelist = []

	# Score value has changed
	if game.score.has_changed() or _force:
		# draw it
		changelist += sprite_draw_occasional( game.score )
		# update the high score if that's changed
		game.hiscore.hiscore( game.score )

	if game.hiscore.has_changed() or _force:
		changelist += sprite_draw_occasional( game.hiscore )

	# number of ships has changed
	if game.ship_count.has_changed() or _force:
		# And draw it
		changelist += sprite_draw_occasional( game.ship_count )

	return changelist


# -----------------------------------------------------------------------------
#
# Our infinite game loop
#
# -----------------------------------------------------------------------------
def main_loop():

	# Set up a timing clock
	clock = pygame.time.Clock()

	# Assume we had window focus
	game.focus = FOCUS_ACTIVE
	
	while 1:

		# -----------------------------------------
		# Timing
		# -----------------------------------------

		# Go at most 30 fps, get how long passed since last tick
		seconds = clock.tick(30)/1000.0

		# -----------------------------------------
		# Look for external events. Mostly keys.
		# -----------------------------------------

		for event in pygame.event.get():

			# --------------------
			# Quit event bails us out, whether it comes from the OS
			# or we pushed it back on the event queue.
			#
			# Or quit on alt-f4
			# --------------------
			if event.type == QUIT:
				return

			if event.type==KEYDOWN and event.key==K_F4 and \
				   (pygame.key.get_mods() & (KMOD_LALT|KMOD_RALT)):
				return
				


			# --------------------
			# pass events to menus if they're active
			# --------------------
			elif game.main_menu.is_active():
				game.main_menu.event( event )
			elif game.ingame_menu.is_active():
				game.ingame_menu.event( event )

			# --------------------
			# Key down is usually what we care about
			# --------------------
			elif event.type == KEYDOWN:

				if game.state == game.STATE_DEMO:
					# If we're in demo these keys activate the menu
					if event.key in (K_SPACE, K_ESCAPE, K_RETURN ):
						game.main_menu.activate()

				elif game.state == game.STATE_GAME:
					# in-game menu
					if event.key == K_ESCAPE:
						sound_pause()
						game.ingame_menu.activate()
					
					# if we're in a game, pass keys to the ship
					elif event.key in game.keys:
						sprite.ship.key_event( event )

			# --------------------
			# Don't care about key up so much, though the ship does for movement
			# --------------------
			elif event.type == KEYUP:
				# if we're in a game, pass keys to the ship
				if (game.state==game.STATE_GAME) and event.key in game.keys:
					sprite.ship.key_event( event )

			# --------------------
			# Active Events determine whether we have focus
			# --------------------
			elif event.type == ACTIVEEVENT:
				focus_activeevent( event )


		# -----------------------------------------
		# The game is running only if we're in the
		# window and not in the ingame menu
		# -----------------------------------------
		was_focused = focus_have_focus()
		running = was_focused and not game.ingame_menu.is_active()

		# Pump our timed event callbacks
		if running:
			TimedEvent.pump( seconds )

		# -----------------------------------------
		# Clear sprites before next frame
		# -----------------------------------------

		sprite_set.demo.clear( game.screen, game.background )
		sprite_set.ship.clear( game.screen, game.background )
		sprite_set.planes.clear( game.screen, game.background )
		sprite_set.harmless.clear( game.screen, game.background )
		sprite_set.clouds.clear( game.screen, game.background )
		sprite_set.bullets.clear( game.screen, game.background )
			
		sprite_set.menu.clear( game.screen, game.background )
		sprite_set.text.clear( game.screen, game.background )

		# -----------------------------------------
		# Do the demo stuff
		# -----------------------------------------
		if running and game.state==game.STATE_DEMO:
			Demo.pump( seconds )

		# -----------------------------------------
		# Collisions!
		# -----------------------------------------

		if running and game.state==game.STATE_GAME:

			# Check for sam colliding with anything
			if sprite.sam:
				sam_check_collisions()

			# Handle for our awful death
			ship_death_logic()

	
		# -----------------------------------------
		# Update the sprites - most only if we have focus
		# -----------------------------------------

		if running:
			# Update waves
			Wave.update_all( seconds )
			# Update the ship position
			sprite_set.ship.update( seconds )

			# These are usually empty, so no harm
			sprite_set.demo.update( seconds )

			# Update the plane marching base position, tell the planes where the
			# ship is at, then update all planes
			if not level.done:
				Plane.update_basepos( seconds )
				# Start patrolling if the ship isn't dying
				if not sprite.ship.is_dying():
					planes_start_patrol( seconds )
					cruise_start_patrol( seconds )
					bomber_start_patrol( seconds )
				Plane.set_ship_pos( [sprite.ship.xpos, sprite.ship.ypos] )
				sprite_set.planes.update( seconds )

			sprite_set.bullets.update( seconds )
			sprite_set.clouds.update( seconds )
			sprite_set.harmless.update( seconds )

		# always update the text - this give us our paused text
		sprite_set.text.update( seconds )
		sprite_set.menu.update( seconds )

		# -----------------------------------------
		# Draw the screen
		# -----------------------------------------

		# Start with empty list of draw changes
		changelist = []

		# Clip to the play area - don't overwrite the UI part
		game.screen.set_clip( PLAY_RECT )

		# Draw all the items we know about
		changelist += Wave.draw_all(game.screen)
		changelist += sprite_set.clouds.draw(game.screen)
		changelist += sprite_set.demo.draw(game.screen)
		changelist += sprite_set.ship.draw(game.screen)
		changelist += sprite_set.planes.draw(game.screen)
		changelist += sprite_set.harmless.draw(game.screen)
		changelist += sprite_set.bullets.draw(game.screen)

		# return clip to normal
		game.screen.set_clip( )

		# draw the overlay stuff
		changelist += sprite_set.menu.draw(game.screen)
		changelist += sprite_set.text.draw(game.screen)

		# Stuff that doesn't change every frame
		changelist += draw_occasional_stuff()

		# Actually update the screen
		pygame.display.update( changelist )

		# -----------------------------------------
		# Are we done with the level?
		# -----------------------------------------

		if running and game.state==game.STATE_GAME and not level.done:
			level_check_for_done()
			
