#
# Turn Wavy Navy into a standalone windows setup
#
# Run this with
#     python -OO bin\package_win.py py2exe -b 2
#
# NOTE: For some reason freesansbold.ttf isn't being found by pygame after
#   packaging - put it in library.zip in the pygame directory


from distutils.core import setup
import glob, os, shutil, sys
import py2exe
import pygame

sys.path += ['.\\bin' ]
from Game import VERSION

# ----------------------------------------------------------------
# Extra files we need to include, and those we want to erase
# ----------------------------------------------------------------

# Need system font
pygamedir = os.path.split(pygame.base.__file__)[0]
font = ( os.path.join(pygamedir, pygame.font.get_default_font()))

# everything else we need to move into dist\bin
keep_in_root = [
	'README.txt', 'issues.txt', 'LICENSE.txt',
	'user_config.txt', 'wavy.bat', font ]

# Our other dist files
extra_data = [
	( "docs", glob.glob("docs/*.*" )),
	( ".", keep_in_root )
	]

for root, dirs, files in os.walk( "skins" ):
	if ".svn" in root: continue
	files = [ os.path.join( root, f ) for f in files ]
	extra_data.append( (root, files) )

# Delete these that py2exe creates - not necessary for wavy
excludes = [
	# windowed version claims these are 'errors' even through they're
	# not necessary, so just leave them in:
	#    'joystick.pyd', 'cdrom.pyd',
	#'movie.pyd', 'surfarray.pyd', 'sndarray.pyd',
	]

if __name__ == "__main__":

	# Wipe out the entire dist tree so we start clean
	print "\n## Wiping old dist tree"
	try:
		shutil.rmtree( "dist" )
	except:
		pass

	print "\n## Running py2exe"
	# Let py2exe do its work
	setup(
		# Description info
		version = VERSION,
		description = "Wavy Navy done in Python and PyGame",
		name = "Wavy Navy",

		# targets
		#console = [ "bin/wavy.py" ],
		windows = [ "bin/wavy.py" ],
	
		options = { 
			"py2exe": {
				"optimize": 2,
				"compressed": 1,
				#"excludes": excludes  # doesn't seem to work
				},
			},
		data_files = extra_data,
		)

	print "\n## Removing extra files"
	# delete the extra files it created that we don't want
	for file in excludes:
		os.remove( os.path.join( "dist", file ) )


	# Now move almost everything in dist into dist/bin
	print "\n## Moving files into dist/bin"
	os.mkdir( "dist/bin" )
	for file in glob.glob("dist/*.*"):
		if os.path.isdir(file):
			continue
		if os.path.split(file)[-1] in keep_in_root:
			continue
		shutil.move( file, "dist/bin" )
	

	print "\n## Wiping old build tree"
	shutil.rmtree( "build" )
